#include <wx/wx.h>
#include <wx/arrstr.h>
#include <wx/filename.h>
#include <wx/textfile.h>
#include <wx/image.h>

#include "FGIModel.h"

#include "ObjConvert.h"
#include "XObjDefs.h"
#include "XObjReadWrite.h"

FGIModel::FGIModel(wxFileName name, wxArrayString srcfiles, wxInt32 xpindex)
{
    Name     = name;
    SrcFiles = srcfiles;
    XPIndex  = xpindex;
    HaveXML  = false;
}

wxString FGIModel::GetName()
{
    wxString retstr;
    retstr = Name.GetFullPath();
    if (HaveXML)
        retstr.Replace(wxT(".obj"), wxT(".xml"));
    else
        retstr.Replace(wxT(".obj"), wxT(".ac"));
    return retstr;
}


wxInt32 FGIModel::GetXPIndex()
{
    return XPIndex;
}

void FGIModel::SetXPIndex(wxInt32 xpindex)
{
    XPIndex = xpindex;
}

wxString FGIModel::FindTexture(wxFileName srcdir, wxFileName modelfile, wxFileName texfile)
{
    bool texfound = false;
    wxString texpath = modelfile.GetPath() + wxFileName::GetPathSeparator() + texfile.GetFullName();
    if (wxFileExists(texpath))
    {
        texfound = true;
    }
    else
    {
        wxString tmpdir = srcdir.GetFullPath() + wxT("custom object textures");
        if (wxDirExists(tmpdir))
        {
            texpath = tmpdir + wxFileName::GetPathSeparator() + texfile.GetFullName();
            if (wxFileExists(texpath))
                texfound = true;
        }
    }

    if (texfound)
        return texpath;

    texpath = modelfile.GetPath() + wxFileName::GetPathSeparator() + texfile.GetName() + wxT(".bmp");
    if (wxFileExists(texpath))
    {
        texfound = true;
    }
    else
    {
        wxString tmpdir = srcdir.GetFullPath() + wxT("custom object textures");
        if (wxDirExists(tmpdir))
        {
            texpath = tmpdir + wxFileName::GetPathSeparator() + texfile.GetName() + wxT(".bmp");
            if (wxFileExists(texpath))
                texfound = true;
        }
    }

    if (!texfound)
        return wxEmptyString;

    wxInitAllImageHandlers();
    wxImage texture;
    texture.LoadFile(texpath);
    texpath.Replace(wxT(".bmp"), wxT(".png"));
    texture.SaveFile(texpath, wxBITMAP_TYPE_PNG);

    return texpath;
}

bool FGIModel::toAC(wxString outdir, wxFileName srcdir)
{
    wxString tmpstr;
    tmpstr = Name.GetFullPath();
    tmpstr.Replace(wxT(".obj"), wxT(".ac"));

    wxFileName filename(outdir + wxFileName::GetPathSeparator() + tmpstr);
    filename.Normalize();

    if (!wxDirExists(filename.GetPath()))
        wxFileName::Mkdir(filename.GetPath(), 0777, wxPATH_MKDIR_FULL);

    if (wxFileExists(filename.GetFullPath()))
    {
        wxLogVerbose(_("Not overwriting %s"), filename.GetFullPath().c_str());
        return true;
    }

    wxTextFile output_file(filename.GetFullPath());
    output_file.Create(filename.GetFullPath());

    output_file.AddLine(wxT("AC3Db"));
    output_file.AddLine(wxT("MATERIAL \"Default\" rgb 0.8 0.8 0.8 amb 0.5 0.5 0.5 emis 0.5 0.5 0.5 spec 1 1 1 shi 32 trans 0"));
    output_file.AddLine(wxT("OBJECT world"));
    output_file.AddLine(wxString::Format(wxT("kids %d"), SrcFiles.GetCount()));
    for (size_t i = 0; i < SrcFiles.GetCount(); i++)
    {
        output_file.AddLine(wxT("OBJECT poly"));
        output_file.AddLine(wxString::Format(wxT("name \"Mesh.%03d\""), i));
        wxArrayString txtobj = ACString(wxFileName(SrcFiles.Item(i)));
        for (size_t j = 0; j < txtobj.GetCount(); j++)
        {
            output_file.AddLine(txtobj.Item(j));
        }
    }
    output_file.AddLine(wxT("kids 0"));
    output_file.Write();
    output_file.Close();

    wxArrayString xmlstr;
    xmlstr.Add(wxT("<?xml version=\"1.0\"?>\n"));
    xmlstr.Add(wxT("<PropertyList>\n"));
    xmlstr.Add(wxString::Format(wxT("  <path>%s</path>\n"), tmpstr.c_str()));

    for (size_t i = 0; i < SrcFiles.GetCount(); i++)
    {
        if (Textures.Item(i) != wxEmptyString)
        {
            wxString texpath = FindTexture(srcdir, SrcFiles.Item(i), Textures.Item(i));
            wxFileName texname(Textures.Item(i));
            if (texpath != wxEmptyString)
                wxCopyFile(texpath, filename.GetPath() + wxFileName::GetPathSeparator() + texname.GetFullName());
        }

        if (Textures_LIT.Item(i) != wxEmptyString)
        {
            wxString texpath = FindTexture(srcdir, SrcFiles.Item(i), Textures_LIT.Item(i));
            wxFileName texname(Textures_LIT.Item(i));
            if (texpath == wxEmptyString)
            {
                wxString litname = Textures_LIT.Item(i);
                litname.Replace(wxT("_LIT"), wxT("LIT"));
                texpath = FindTexture(srcdir, SrcFiles.Item(i), litname);
            }

            if (texpath != wxEmptyString)
                wxCopyFile(texpath, filename.GetPath() + wxFileName::GetPathSeparator() + texname.GetFullName());
        }
    }

    xmlstr.Add(wxT("</PropertyList>"));

    HaveXML = true;
    tmpstr.Replace(wxT(".ac"), wxT(".xml"));
    wxFileName xmlname(outdir + wxFileName::GetPathSeparator() + tmpstr);
    xmlname.Normalize();
    wxTextFile xml_file(xmlname.GetFullPath());
    xml_file.Create(xmlname.GetFullPath());
    for (size_t i = 0; i < xmlstr.GetCount(); i++)
    {
        xml_file.AddLine(xmlstr.Item(i));
    }
    xml_file.Write();
    xml_file.Close();

    return true;
}

wxArrayString FGIModel::ACString(wxFileName infile)
{
    infile.Normalize();
    wxArrayString tmpstr;
    wxArrayString retstr;

    if (!wxFileExists(infile.GetFullPath()))
    {
        wxLogWarning(_("Cannot find model %s"), infile.GetFullPath().c_str());
        return retstr;
    }

    wxLogMessage(_("Converting model %s"), infile.GetFullPath().c_str());

    XObj model7;
    XObj8 model;
    if (XObjRead(infile.GetFullPath().mb_str(), model7))
        Obj7ToObj8(model7, model);
    else if (!XObj8Read(infile.GetFullPath().mb_str(), model))
        return retstr;

    wxString tex(model.texture.c_str(), wxConvUTF8);
    tex.Replace(wxT("\\"), wxT("/"), true);
    tex.Replace(wxT(":"), wxT("/"), true);
    Textures.Add(tex);
    if (!model.texture.empty()) {
        retstr.Add(wxString::Format(wxT("texture \"%s\""), tex.c_str()));
    }

    wxString tex_lit(model.texture_lit.c_str(), wxConvUTF8);
    tex_lit.Replace(wxT("\\"), wxT("/"), true);
    tex_lit.Replace(wxT(":"), wxT("/"), true);
    Textures_LIT.Add(tex_lit);

    ObjPointPool vts;

    const float* v;
    const float* w;

    wxInt32 tris;
    bool smooth = true;
    bool twosided = false;
    XObjLOD8 lod = model.lods[0];
    for (vector<XObjCmd8>::const_iterator cmd = lod.cmds.begin(); cmd != lod.cmds.end(); ++cmd)
    {
        switch(cmd->cmd) {
            case attr_Shade_Smooth:
                smooth = true;
                break;

            case attr_Shade_Flat:
                smooth = false;
                break;

            case attr_NoCull:
                twosided = true;
                break;

            case attr_Cull:
                twosided = false;
                break;

            case obj8_Tris:
                tris = cmd->idx_count / 3;
                tmpstr.Add(wxString::Format(wxT("numsurf %d"), tris));
                for (size_t i = 0; i < tris; i++)
                {
                    wxInt32 surf_type = 0;
                    if (smooth)
                        surf_type += 1;
                    if (twosided)
                        surf_type += 2;
                    tmpstr.Add(wxString::Format(wxT("SURF 0x%d0"), surf_type));
                    tmpstr.Add(wxT("mat 0"));
                    tmpstr.Add(wxT("refs 3"));
                    for (size_t j = 0; j < 3; j++)
                    {
                        wxInt32 num = i * 3 + 2 - j + cmd->idx_offset;
                        wxInt32 pos = model.indices[num];
                        v = model.geo_tri.get(pos);
                        wxInt32 idx;
                        bool dup = false;
                        for (int m = 0; m < vts.count(); m++)
                        {
                            const float* w = vts.get(m);
                            if ((m != pos) && (w[0] == v[0]) && (w[1] == v[1]) && (w[2] == v[2]))
                            {
                                dup = true;
                                idx = m;
                                break;
                            }
                        }
                        if (!dup)
                        {
                            vts.append(v);
                            idx = vts.count() - 1;
                        }
                        tmpstr.Add(wxString::Format(wxT("%d %f %f"), idx, v[6], v[7]));
                    }
                }
                break;
        }
    }

    wxInt32 verts = vts.count();
    retstr.Add(wxString::Format(wxT("numvert %d"), verts));
    for (size_t i = 0; i < vts.count(); i++)
    {
        v = vts.get(i);
        retstr.Add(wxString::Format(wxT("%f %f %f"), v[0], v[1], v[2]));
    }

    for (size_t i = 0; i < tmpstr.GetCount(); i++)
    {
        retstr.Add(tmpstr.Item(i));
    }

    return retstr;
}

