#include <wx/wx.h>
#include <wx/dir.h>
#include <wx/arrstr.h>

#include "FGIScenery.h"
#include "FGITile.h"
#include "FGIModel.h"

#include "DSFLib.h"

#include <wx/arrimpl.cpp>
WX_DEFINE_OBJARRAY(FGITileArray);

#include <wx/arrimpl.cpp>
WX_DEFINE_OBJARRAY(FGIModelArray);

FGIScenery::FGIScenery(wxString name, wxFileName path, wxFileName output, bool static_mod, bool group_mod)
{
    Name         = name;
    Path         = path;
    Type         = SceneryUnknown;
    OutputDir    = output;
    StaticModels = static_mod;
    GroupModels  = group_mod;
}

bool FGIScenery::Found()
{
    wxString tmpdir = Path.GetFullPath() + wxT("Earth nav data");

    if (!wxDirExists(tmpdir))
        return false;

    wxArrayString tmpfiles;
    if (wxDir::GetAllFiles(tmpdir, &tmpfiles, wxT("*.dsf"), wxDIR_DIRS | wxDIR_FILES) > 0)
    {
        Type = SceneryXPDsf;
        return true;
    }

    if (wxDir::GetAllFiles(tmpdir, &tmpfiles, wxT("*.env"), wxDIR_DIRS | wxDIR_FILES) > 0)
    {
        Type = SceneryXPEnv;
        return true;
    }

    return false;
}

bool FGIScenery::Load()
{
    wxArrayString tiles;

    switch (Type)
    {
    case SceneryXPDsf:
        wxString datadir = Path.GetFullPath() + wxT("Earth nav data");
        wxDir::GetAllFiles(datadir, &tiles, wxT("*.dsf"), wxDIR_DIRS | wxDIR_FILES);
        for (size_t i = 0; i < tiles.GetCount(); ++i)
        {
            wxLogMessage(_("Loading DSF file %s"), tiles.Item(i).c_str());
            CurrentModel = 0;
            CleanIndexes();
            LoadDSF(tiles.Item(i));
            // **** Calculate elevations
        }
    }

    return true;
}

void FGIScenery::CleanIndexes()
{
    for (size_t i = 0; i < Models.GetCount(); i++)
    {
        Models.Item(i).SetXPIndex(-1);
    }
}

FGITile* FGIScenery::GetTile(wxString name, wxString path)
{
    if (Tiles.IsEmpty())
    {
        return NULL;
    }

    for (size_t i = 0; i < Tiles.GetCount(); i++)
    {
        if ((Tiles.Item(i).GetName() == name) && (Tiles.Item(i).GetPath() == path))
        {
            return &Tiles.Item(i);
        }
    }

    return NULL;
}

wxString FGIScenery::GetName()
{
    return Name;
}

bool FGIScenery::IsStatic()
{
    return StaticModels;
}

wxInt32 FGIScenery::GetModelFromXPIndex(wxInt32 xpindex)
{
    for (size_t i = 0; i < Models.GetCount(); i++)
    {
        if (Models.Item(i).GetXPIndex() == xpindex)
            return i;
    }

    return -1;
}

/*
 * Declare DSFLib callbacks
 *
 */

/* Fake functions we will not use */
void fake_AcceptTerrainDef(const char * inPartialPath, void * inRef) {}
void fake_AcceptPolygonDef(const char * inPartialPath, void * inRef) {}
void fake_AcceptNetworkDef(const char * inPartialPath, void * inRef) {}
void fake_AcceptProperty(const char * inProp, const char * inValue, void * inRef) {}
void fake_BeginPatch(unsigned int inTerrainType, double inNearLOD, double inFarLOD, unsigned char inFlags, int inCoordDepth, void * inRef) {}
void fake_BeginPrimitive(int inType, void * inRef) {}
void fake_AddPatchVertex(double inCoordinates[], void * inRef) {}
void fake_EndPrimitive(void * inRef) {}
void fake_EndPatch(void * inRef) {}
void fake_BeginSegment(unsigned int inNetworkType, unsigned int inNetworkSubtype, unsigned int inStartNodeID, double inCoordinates[6], bool inCurved, void * inRef) {}
void fake_AddSegmentShapePoint(double inCoordinates[6], bool inCurved, void * inRef) {}	
void fake_EndSegment(unsigned int inEndNodeID, double inCoordinates[6], bool inCurved, void * inRef) {}					
void fake_BeginPolygon(unsigned int inPolygonType, unsigned short inParam, int inDepth, void * inRef) {}
void fake_BeginPolygonWinding(void * inRef) {}
void fake_AddPolygonPoint(double inCoordinates[2], void * inRef) {}
void fake_EndPolygonWinding(void * inRef) {}
void fake_EndPolygon(void * inRef) {}

/* We don't want DSFReadFile to go on forever */
bool FGIScenery::DSFNextPass(int finished_pass_index, void * inRef)
{
    return true;
}

void FGIScenery::DSFAcceptObjectDef(const char * inPartialPath, void * inRef)
{
    FGIScenery* myScenery = (FGIScenery*) inRef;

    wxString path(inPartialPath, wxConvUTF8);
    path.Replace(wxT("\\"), wxT("/"), true);
    path.Replace(wxT(":"), wxT("/"), true);
    wxLogVerbose(_("Got object %s"), path.c_str());

    wxFileName modelfile(path);
    wxFileName modelpath(myScenery->Path.GetFullPath() + modelfile.GetFullPath());

    bool modelfound = false;

    /* A model can be in the base directory or in a 'custom objects' directory */
    if (wxFileExists(modelpath.GetFullPath()))
    {
        modelfound = true;
    }
    else
    {
        wxString tmpdir = myScenery->Path.GetFullPath() + wxT("custom objects");
        if (wxDirExists(tmpdir))
        {
            modelpath.Assign(tmpdir + wxFileName::GetPathSeparator() + modelfile.GetFullPath());
            if (wxFileExists(modelpath.GetFullPath()))
                modelfound = true;
        }
    }
    
    if (modelfound)
    {
        wxLogVerbose(_("Found object %s"), modelpath.GetFullPath().c_str());

        wxArrayString srcfiles;
        srcfiles.Add(modelpath.GetFullPath());

        wxString modelname = modelfile.GetFullPath();
        modelname.Replace(wxT(" "), wxT("_"));

        if (myScenery->IsStatic())
        {
            FGIModel newModel(modelname, srcfiles, myScenery->CurrentModel);
            myScenery->Models.Add(newModel);
        }
        else
        {
            FGIModel newModel(wxT("Models/") + myScenery->GetName() + wxT("/") + modelname, srcfiles, myScenery->CurrentModel);
            myScenery->Models.Add(newModel);
        }
    }
    else
    {
        wxLogWarning(_("Cannot find model %s"), path.c_str());
    }

    myScenery->CurrentModel++;
}

void FGIScenery::DSFAddObject(unsigned int inObjectType, double inCoordinates[2], double inRotation, void * inRef)
{
    FGIScenery* myScenery = (FGIScenery*) inRef;

    wxString name = FGITile::GetNameFromDeg(inCoordinates[0], inCoordinates[1]);
    wxString path = FGITile::GetPathFromDeg(inCoordinates[0], inCoordinates[1]);

    wxLogVerbose(_("Got object placement"));

    wxString lon = wxString::Format(wxT("%lf"), inCoordinates[0]);
    lon.Replace(wxT(","), wxT("."), true);
    wxString lat = wxString::Format(wxT("%lf"), inCoordinates[1]);
    lat.Replace(wxT(","), wxT("."), true);

    wxDouble alt;
    alt = -9999;

    wxInt32 ModelIndex = myScenery->GetModelFromXPIndex(inObjectType);
    FGIObject newObject(ModelIndex, inCoordinates[0], inCoordinates[1], alt, 90 - inRotation);
    FGITile* myTile = myScenery->GetTile(name, path);

    if (myTile == NULL)
    {
        wxLogVerbose(_("Creating tile %s/%s.stg"), path.c_str(), name.c_str());
        FGITile newTile(name, path);
        newTile.Objects.Add(newObject);
        myScenery->Tiles.Add(newTile);
    }
    else
    {
        wxLogVerbose(_("Adding object to %s/%s.stg"), path.c_str(), name.c_str());
        myTile->Objects.Add(newObject);
    }
}

bool FGIScenery::LoadDSF(wxString name)
{
    DSFCallbacks_t cbs;

    cbs.AcceptTerrainDef_f     = fake_AcceptTerrainDef;
    cbs.AcceptPolygonDef_f     = fake_AcceptPolygonDef;
    cbs.AcceptNetworkDef_f     = fake_AcceptNetworkDef;
    cbs.AcceptProperty_f       = fake_AcceptProperty;
    cbs.BeginPatch_f           = fake_BeginPatch;
    cbs.BeginPrimitive_f       = fake_BeginPrimitive;
    cbs.AddPatchVertex_f       = fake_AddPatchVertex;
    cbs.EndPrimitive_f         = fake_EndPrimitive;
    cbs.EndPatch_f             = fake_EndPatch;
    cbs.BeginSegment_f         = fake_BeginSegment;
    cbs.AddSegmentShapePoint_f = fake_AddSegmentShapePoint;
    cbs.EndSegment_f           = fake_EndSegment;
    cbs.BeginPolygon_f         = fake_BeginPolygon;
    cbs.BeginPolygonWinding_f  = fake_BeginPolygonWinding;
    cbs.AddPolygonPoint_f      = fake_AddPolygonPoint;
    cbs.EndPolygonWinding_f    = fake_EndPolygonWinding;
    cbs.EndPolygon_f           = fake_EndPolygon;

    cbs.AcceptObjectDef_f      = FGIScenery::DSFAcceptObjectDef;
    cbs.AddObject_f            = FGIScenery::DSFAddObject;
    cbs.NextPass_f             = FGIScenery::DSFNextPass;

    DSFReadFile(name.mb_str(), &cbs, NULL, (void*) this);

    return true;
}

