#include <wx/wx.h>
#include <wx/cmdline.h>
#include <wx/filename.h>
#include <wx/textfile.h>
#include <wx/tokenzr.h>

#include <simgear/bucket/newbucket.hxx>      // for SGBucket
#include <simgear/scene/tgdb/obj.hxx>        // for SGLoadBTG
#include <simgear/scene/material/matlib.hxx> // for SGMaterialLib
#include <simgear/misc/sgstream.hxx>
#include <osgUtil/IntersectVisitor>

#include "FGIScenery.h"
#include "FGITile.h"
#include "FGIObject.h"
#include "FGIModel.h"

class FGImport: public wxAppConsole
{
private:
    wxString   Name;
    wxFileName Path;
    wxFileName OutputDir;
    wxFileName FGPath;

    bool       StaticModels;
    bool       GroupModels;

    void       OnInitCmdLine(wxCmdLineParser& parser);
    bool       OnCmdLineParsed(wxCmdLineParser& parser);
    int        OnRun();
};

/* Declare application */
IMPLEMENT_APP(FGImport)

wxDouble GetElev(osg::Group* rootNode, wxDouble lon, wxDouble lat)
{
    // Create intersector segment
    SGVec3d start = SGVec3d::fromGeod(SGGeod::fromDegM(lon, lat, 9999));
    SGVec3d end = SGVec3d::fromGeod(SGGeod::fromDegM(lon, lat, -9999));

    // Create intersector
    osgUtil::IntersectVisitor intersectVisitor;
    osg::LineSegment* lineSegment = new osg::LineSegment(start.osg(), end.osg());
    intersectVisitor.addLineSegment(lineSegment);
    rootNode->accept(intersectVisitor);

    // Get intersections
    bool hits = intersectVisitor.hits();
    double alt = -9999;
    if (hits) {
        int nHits = intersectVisitor.getNumHits(lineSegment);
        for (int i = 0; i < nHits; ++i) {
            const osgUtil::Hit& hit = intersectVisitor.getHitList(lineSegment)[i];
            SGVec3d point;
            point.osg() = hit.getWorldIntersectPoint();
            SGGeod geod = SGGeod::fromCart(point);
            double elevation = geod.getElevationM();
            if (alt < elevation) {
                alt = elevation;
            }
        }
    }

    return alt;
}

void FGImport::OnInitCmdLine(wxCmdLineParser& parser)
{
    /* Call parent function to get help and verbose command line options */
    wxAppConsole::OnInitCmdLine(parser);

    static const wxCmdLineEntryDesc cmdLineDesc[] =
    {
        { wxCMD_LINE_OPTION, wxT("n"),      wxT("name"),       wxT("name of the scenery (name of the output directory if not specified)") },
        { wxCMD_LINE_SWITCH, wxT("s"),      wxT("static"),     wxT("make objects static") },
        { wxCMD_LINE_SWITCH, wxT("g"),      wxT("group"),      wxT("group objects that are at the same coordinates") },
        { wxCMD_LINE_OPTION, wxT("p"),      wxT("fg-scenery"), wxT("FG scenery path") },
        { wxCMD_LINE_PARAM,  wxEmptyString, wxEmptyString,     wxT("<input dir>") },
        { wxCMD_LINE_PARAM,  wxEmptyString, wxEmptyString,     wxT("<output dir>") },
        { wxCMD_LINE_NONE }
    };

    parser.SetDesc(cmdLineDesc);
}

bool FGImport::OnCmdLineParsed(wxCmdLineParser& parser)
{
    /* Call parent function to process help and verbose command line options */
    wxAppConsole::OnCmdLineParsed(parser);

    Path.AssignDir(parser.GetParam(0));
    Path.Normalize();
    if (!wxDirExists(Path.GetFullPath()))
        wxLogFatalError(_("Directory %s does not exist"), parser.GetParam(0).c_str());

    OutputDir.AssignDir(parser.GetParam(1));
    OutputDir.Normalize();
    if (wxDirExists(parser.GetParam(1)))
        wxLogWarning(_("Overwriting scenery files in %s but NOT overwriting models"), OutputDir.GetFullPath().c_str());

    if (!parser.Found(wxT("n"), &Name))
    {
        wxArrayString dirs = OutputDir.GetDirs();
        Name = dirs[OutputDir.GetDirCount() - 1];
    }

    wxString PathArg = wxEmptyString;
    if (parser.Found(wxT("p"), &PathArg))
    {
        FGPath = wxFileName(PathArg);
        FGPath.Normalize();
    }

    if (parser.Found(wxT("s")))
        StaticModels = true;
    else
        StaticModels = false;

    if (parser.Found(wxT("g")))
        GroupModels = true;
    else
        GroupModels = false;
}

int FGImport::OnRun()
{
    FGIScenery Scenery(Name, Path, OutputDir, StaticModels, GroupModels);

    if (!Scenery.Found())
        wxLogFatalError(_("No scenery found in %s"), Path.GetFullPath().c_str());

    wxLogMessage(_("Found scenery in %s"), Path.GetFullPath().c_str());

    Scenery.Load();

    FGITile* currentTile;
    wxString outdir;
    FGIObject* currentObject;

    // Create root node of the scene
    osg::Group* rootNode = new osg::Group();
    SGMaterialLib matlib;

    for (size_t i = 0; i < Scenery.Tiles.GetCount(); i++)
    {
        currentTile = &Scenery.Tiles.Item(i);
        wxLogMessage(_("Processing tile %s/%s.stg"), currentTile->GetPath().c_str(), currentTile->GetName().c_str());

        if (StaticModels)
            outdir = OutputDir.GetFullPath() + currentTile->GetPath();
        else
            outdir = OutputDir.GetFullPath() + wxT("Objects") + wxFileName::GetPathSeparator() + currentTile->GetPath();

        if (!wxDirExists(outdir))
            wxFileName::Mkdir(outdir, 0777, wxPATH_MKDIR_FULL);
        
        wxString tilename = outdir + wxFileName::GetPathSeparator() + currentTile->GetName() + wxT(".stg");
        wxTextFile tile_file(tilename);
        tile_file.Create(tilename);

        FGIObject* object1;
        FGIObject* object2;

        // Load stg file (Scene Terra Gear)
        wxString wxbase_path = FGPath.GetFullPath() + wxT("/Terrain/") + currentTile->GetPath();
        string base_path = (const char*)wxbase_path.mb_str(wxConvUTF8);
        string stg_name = base_path + "/" + (const char*)currentTile->GetName().mb_str(wxConvUTF8) + ".stg";
        sg_gzifstream in(stg_name);
        while (!in.eof()) {
            string token;
            in >> token;
            // Load only terrain objects
            if ((token == "OBJECT_BASE") || (token == "OBJECT")) {
                string name;
                in >> name >> ::skipws;
                string obj_name = base_path + "/" + name + ".gz";
                rootNode->addChild(SGLoadBTG(obj_name, &matlib, false, false, false));
            }
        }

        for (size_t j = 0; j < currentTile->Objects.GetCount(); j++)
        {
            currentObject = &currentTile->Objects.Item(j);
            wxDouble alt = GetElev(rootNode, currentObject->GetLon(), currentObject->GetLat());
            currentObject->SetAlt(alt);
            wxInt32 modelid = currentObject->GetModelId();
            if ((modelid >= 0) && (modelid < Scenery.Models.GetCount()))
            {
                if (StaticModels)
                    Scenery.Models.Item(modelid).toAC(outdir, Path);
                else
                    Scenery.Models.Item(modelid).toAC(OutputDir.GetFullPath(), Path);
                wxString modelname = Scenery.Models.Item(modelid).GetName();
                tile_file.AddLine(currentObject->GetDef(modelname, StaticModels));
            }
            else
            {
                wxLogWarning(_("Invalid model id %d"), modelid);
            }
        }

        // Clear rootNode for next pass
        rootNode->removeChildren(0, rootNode->getNumChildren());

        tile_file.Write();
        tile_file.Close();
    }

    return 0;
}

